@@ -2,16 +2,14 @@
 import json
 import threading
 from funcy import memoize, post_processing, ContextDecorator
-from django.db.models.expressions import F
-# Since Django 1.8, `ExpressionNode` is `Expression`
-try:
-    from django.db.models.expressions import ExpressionNode as Expression
-except ImportError:
-    from django.db.models.expressions import Expression
+from django.db import DEFAULT_DB_ALIAS
+from django.db.models.expressions import F, Expression
 
 from .conf import settings
-from .utils import non_proxy, NOT_SERIALIZED_FIELDS
+from .utils import NOT_SERIALIZED_FIELDS
+from .sharding import get_prefix
 from .redis import redis_client, handle_connection_failure, load_script
+from .signals import cache_invalidated
 from .transaction import queue_when_in_transaction
 
 
@@ -20,27 +18,29 @@ __all__ = ('invalidate_obj', 'invalidate_model', 'invalidate_all', 'no_invalidat
 
 @queue_when_in_transaction
 @handle_connection_failure
-def invalidate_dict(model, obj_dict):
+def invalidate_dict(model, obj_dict, using=DEFAULT_DB_ALIAS):
     if no_invalidation.active or not settings.CACHEOPS_ENABLED:
         return
-    model = non_proxy(model)
-    load_script('invalidate')(args=[
+    model = model._meta.concrete_model
+    prefix = get_prefix(_cond_dnfs=[(model._meta.db_table, list(obj_dict.items()))], dbs=[using])
+    load_script('invalidate')(keys=[prefix], args=[
         model._meta.db_table,
         json.dumps(obj_dict, default=str)
     ])
+    cache_invalidated.send(sender=model, obj_dict=obj_dict)
 
 
-def invalidate_obj(obj):
+def invalidate_obj(obj, using=DEFAULT_DB_ALIAS):
     """
     Invalidates caches that can possibly be influenced by object
     """
-    model = non_proxy(obj.__class__)
-    invalidate_dict(model, get_obj_dict(model, obj))
+    model = obj.__class__._meta.concrete_model
+    invalidate_dict(model, get_obj_dict(model, obj), using=using)
 
 
 @queue_when_in_transaction
 @handle_connection_failure
-def invalidate_model(model):
+def invalidate_model(model, using=DEFAULT_DB_ALIAS):
     """
     Invalidates all caches for given model.
     NOTE: This is a heavy artillery which uses redis KEYS request,
@@ -48,19 +48,23 @@ def invalidate_model(model):
     """
     if no_invalidation.active or not settings.CACHEOPS_ENABLED:
         return
-    model = non_proxy(model)
-    conjs_keys = redis_client.keys('conj:%s:*' % model._meta.db_table)
+    model = model._meta.concrete_model
+    # NOTE: if we use sharding dependent on DNF then this will fail,
+    #       which is ok, since it's hard/impossible to predict all the shards
+    prefix = get_prefix(tables=[model._meta.db_table], dbs=[using])
+    conjs_keys = redis_client.keys('%sconj:%s:*' % (prefix, model._meta.db_table))
     if conjs_keys:
         cache_keys = redis_client.sunion(conjs_keys)
         redis_client.delete(*(list(cache_keys) + conjs_keys))
+    cache_invalidated.send(sender=model, obj_dict=None)
 
 
-@queue_when_in_transaction
 @handle_connection_failure
 def invalidate_all():
     if no_invalidation.active or not settings.CACHEOPS_ENABLED:
         return
     redis_client.flushdb()
+    cache_invalidated.send(sender=None, obj_dict=None)
 
 
 class InvalidationState(threading.local):
