@@ -3,8 +3,8 @@ import sys
 import json
 import threading
 import six
-from funcy import select_keys, cached_property, once, once_per, monkey, wraps, walk
-from funcy.py2 import mapcat, map
+from funcy import select_keys, cached_property, once, once_per, monkey, wraps, walk, chain
+from funcy.py3 import lmap, map, lcat, join_with
 from .cross import pickle, md5
 
 import django
@@ -14,18 +14,14 @@ from django.db.models import Manager, Model
 from django.db.models.query import QuerySet
 from django.db.models.sql.datastructures import EmptyResultSet
 from django.db.models.signals import pre_save, post_save, post_delete, m2m_changed
-# This thing was removed in Django 1.8
-try:
-    from django.db.models.query import MAX_GET_RESULTS
-except ImportError:
-    MAX_GET_RESULTS = None
 
-from .conf import model_profile, model_is_fake, settings, ALL_OPS
+from .conf import model_profile, settings, ALL_OPS
 from .utils import monkey_mix, stamp_fields, func_cache_key, cached_view_fab, family_has_profile
+from .sharding import get_prefix
 from .redis import redis_client, handle_connection_failure, load_script
 from .tree import dnfs
 from .invalidation import invalidate_obj, invalidate_dict, no_invalidation
-from .transaction import transaction_state
+from .transaction import transaction_states
 from .signals import cache_read
 
 
@@ -35,13 +31,15 @@ _local_get_cache = {}
 
 
 @handle_connection_failure
-def cache_thing(cache_key, data, cond_dnfs, timeout):
+def cache_thing(prefix, cache_key, data, cond_dnfs, timeout, dbs=()):
     """
     Writes data to cache and creates appropriate invalidators.
     """
-    assert not transaction_state.is_dirty()
+    # Could have changed after last check, sometimes superficially
+    if transaction_states.is_dirty(dbs):
+        return
     load_script('cache_thing', settings.CACHEOPS_LRU)(
-        keys=[cache_key],
+        keys=[prefix, cache_key],
         args=[
             pickle.dumps(data, -1),
             json.dumps(cond_dnfs, default=str),
@@ -59,6 +57,8 @@ def cached_as(*samples, **kwargs):
     extra = kwargs.pop('extra', None)
     key_func = kwargs.pop('key_func', func_cache_key)
     lock = kwargs.pop('lock', None)
+    if not samples:
+        raise TypeError('Pass a queryset, a model or an object to cache like')
     if kwargs:
         raise TypeError('Unexpected keyword arguments %s' % ', '.join(kwargs))
 
@@ -79,11 +79,12 @@ def cached_as(*samples, **kwargs):
 
         return queryset
 
-    querysets = map(_get_queryset, samples)
-    cond_dnfs = mapcat(dnfs, querysets)
-    key_extra = [qs._cache_key() for qs in querysets]
+    querysets = lmap(_get_queryset, samples)
+    dbs = list({qs.db for qs in querysets})
+    cond_dnfs = join_with(lcat, map(dnfs, querysets))
+    key_extra = [qs._cache_key(prefix=False) for qs in querysets]
     key_extra.append(extra)
-    if not timeout:  # TODO: switch to is None on major release
+    if timeout is None:
         timeout = min(qs._cacheprofile['timeout'] for qs in querysets)
     if lock is None:
         lock = any(qs._cacheprofile['lock'] for qs in querysets)
@@ -91,10 +92,11 @@ def cached_as(*samples, **kwargs):
     def decorator(func):
         @wraps(func)
         def wrapper(*args, **kwargs):
-            if transaction_state.is_dirty() or not settings.CACHEOPS_ENABLED:
+            if not settings.CACHEOPS_ENABLED or transaction_states.is_dirty(dbs):
                 return func(*args, **kwargs)
 
-            cache_key = 'as:' + key_func(func, args, kwargs, key_extra)
+            prefix = get_prefix(func=func, _cond_dnfs=cond_dnfs, dbs=dbs)
+            cache_key = prefix + 'as:' + key_func(func, args, kwargs, key_extra)
 
             with redis_client.getting(cache_key, lock=lock) as cache_data:
                 cache_read.send(sender=None, func=func, hit=cache_data is not None)
@@ -102,7 +104,7 @@ def cached_as(*samples, **kwargs):
                     return pickle.loads(cache_data)
                 else:
                     result = func(*args, **kwargs)
-                    cache_thing(cache_key, result, cond_dnfs, timeout)
+                    cache_thing(prefix, cache_key, result, cond_dnfs, timeout, dbs=dbs)
                     return result
 
         return wrapper
@@ -131,7 +133,7 @@ class QuerySetMixin(object):
                 'you can configure it with empty ops.'
                     % (self.model._meta.app_label, self.model._meta.model_name))
 
-    def _cache_key(self):
+    def _cache_key(self, prefix=True):
         """
         Compute a cache key for this queryset
         """
@@ -162,19 +164,27 @@ class QuerySetMixin(object):
         if hasattr(self, 'flat'):
             md.update(str(self.flat))
 
-        return 'q:%s' % md.hexdigest()
+        cache_key = 'q:%s' % md.hexdigest()
+        return self._prefix + cache_key if prefix else cache_key
+
+    @cached_property
+    def _prefix(self):
+        return get_prefix(_queryset=self)
+
+    @cached_property
+    def _cond_dnfs(self):
+        return dnfs(self)
 
     def _cache_results(self, cache_key, results):
-        cond_dnfs = dnfs(self)
-        cache_thing(cache_key, results, cond_dnfs, self._cacheprofile['timeout'])
+        cache_thing(self._prefix, cache_key, results,
+                    self._cond_dnfs, self._cacheprofile['timeout'], dbs=[self.db])
 
-    def cache(self, ops=None, timeout=None, write_only=None, lock=None):
+    def cache(self, ops=None, timeout=None, lock=None):
         """
         Enables caching for given ops
             ops        - a subset of {'get', 'fetch', 'count', 'exists'},
                          ops caching to be turned on, all enabled by default
             timeout    - override default cache timeout
-            write_only - don't try fetching from cache, still write result there
             lock       - use lock to prevent dog-pile effect
 
         NOTE: you actually can disable caching by omiting corresponding ops,
@@ -190,8 +200,6 @@ class QuerySetMixin(object):
 
         if timeout is not None:
             self._cacheprofile['timeout'] = timeout
-        if write_only is not None:
-            self._cacheprofile['write_only'] = write_only
         if lock is not None:
             self._cacheprofile['lock'] = lock
 
@@ -223,12 +231,11 @@ class QuerySetMixin(object):
                 return self
 
         def clone(self, **kwargs):
-            # NOTE: need to copy profile so that clone changes won't affect this queryset
-            if '_cacheprofile' in self.__dict__ and self._cacheprofile:
-                kwargs.setdefault('_cacheprofile', self._cacheprofile.copy())
-
             clone = self._no_monkey._clone(self, **kwargs)
             clone._cloning = self._cloning - 1 if self._cloning else 0
+            # NOTE: need to copy profile so that clone changes won't affect this queryset
+            if self.__dict__.get('_cacheprofile'):
+                clone._cacheprofile = self._cacheprofile.copy()
             return clone
     else:
         def _clone(self, klass=None, setup=False, **kwargs):
@@ -256,55 +263,34 @@ class QuerySetMixin(object):
             clone._cloning = self._cloning - 1 if self._cloning else 0
             return clone
 
-    def iterator(self):
-        # TODO: drop this in next major release
-        # If cache is not enabled or in transaction just fall back
-        if not self._cacheprofile or 'fetch' not in self._cacheprofile['ops'] \
-                or transaction_state.is_dirty() or not settings.CACHEOPS_ENABLED:
-            return self._no_monkey.iterator(self)
-
-        cache_key = self._cache_key()
-        if not self._cacheprofile['write_only'] and not self._for_write:
-            # Trying get data from cache
-            cache_data = redis_client.get(cache_key)
-            cache_read.send(sender=self.model, func=None, hit=cache_data is not None)
-            if cache_data is not None:
-                return iter(pickle.loads(cache_data))
-
-        # Cache miss - fetch data from overriden implementation
-        def iterate():
-            # NOTE: we are using self._result_cache to avoid fetching-while-fetching bug #177
-            self._result_cache = []
-            for obj in self._no_monkey.iterator(self):
-                self._result_cache.append(obj)
-                yield obj
-            self._cache_results(cache_key, self._result_cache)
-
-        return iterate()
-
     def _fetch_all(self):
-        # If cache is not enabled or in transaction just fall back
-        if not self._cacheprofile or 'fetch' not in self._cacheprofile['ops'] \
-                or transaction_state.is_dirty() or not settings.CACHEOPS_ENABLED:
+        # If already fetched, cache not enabled, within write or in dirty transaction then fall back
+        if self._result_cache is not None \
+                or not settings.CACHEOPS_ENABLED \
+                or not self._cacheprofile or 'fetch' not in self._cacheprofile['ops'] \
+                or self._for_write \
+                or transaction_states[self.db].is_dirty():
             return self._no_monkey._fetch_all(self)
 
-        if self._result_cache is None:
-            cache_key = self._cache_key()
-            lock = self._cacheprofile['lock']
+        cache_key = self._cache_key()
+        lock = self._cacheprofile['lock']
 
-            if self._cacheprofile['write_only'] or self._for_write:
-                self._result_cache = list(self._no_monkey.iterator(self))
-                self._cache_results(cache_key, self._result_cache)
+        with redis_client.getting(cache_key, lock=lock) as cache_data:
+            cache_read.send(sender=self.model, func=None, hit=cache_data is not None)
+            if cache_data is not None:
+                self._result_cache = pickle.loads(cache_data)
             else:
-                with redis_client.getting(cache_key, lock=lock) as cache_data:
-                    cache_read.send(sender=self.model, func=None, hit=cache_data is not None)
-                    if cache_data is not None:
-                        self._result_cache = pickle.loads(cache_data)
-                    else:
-                        self._result_cache = list(self._no_monkey.iterator(self))
-                        self._cache_results(cache_key, self._result_cache)
+                # This thing appears in Django 1.9.
+                # In Djangos 1.9 and 1.10 both calls mean the same.
+                # Starting from Django 1.11 .iterator() uses chunked fetch
+                # while ._fetch_all() stays with bare _iterable_class.
+                if hasattr(self, '_iterable_class'):
+                    self._result_cache = list(self._iterable_class(self))
+                else:
+                    self._result_cache = list(self.iterator())
+                self._cache_results(cache_key, self._result_cache)
 
-        self._no_monkey._fetch_all(self)
+        return self._no_monkey._fetch_all(self)
 
     def count(self):
         if self._cacheprofile and 'count' in self._cacheprofile['ops']:
@@ -316,14 +302,31 @@ class QuerySetMixin(object):
         else:
             return self._no_monkey.count(self)
 
+    def aggregate(self, *args, **kwargs):
+        if self._cacheprofile and 'aggregate' in self._cacheprofile['ops']:
+            # Annotate add all the same annotations, but doesn't run the query
+            qs = self.clone().annotate(*args, **kwargs)
+            return cached_as(qs)(lambda: self._no_monkey.aggregate(self, *args, **kwargs))()
+        else:
+            return self._no_monkey.aggregate(self, *args, **kwargs)
+
     def get(self, *args, **kwargs):
-        # .get() uses the same .iterator() method to fetch data,
+        # .get() uses the same ._fetch_all() method to fetch data,
         # so here we add 'fetch' to ops
         if self._cacheprofile and 'get' in self._cacheprofile['ops']:
             # NOTE: local_get=True enables caching of simple gets in local memory,
             #       which is very fast, but not invalidated.
             # Don't bother with Q-objects, select_related and previous filters,
             # simple gets - thats what we are really up to here.
+            #
+            # TODO: this checks are far from adequete, at least these are missed:
+            #       - settings.CACHEOPS_ENABLED
+            #       - self._for_write
+            #       - self._fields (values, values_list)
+            #       - annotations
+            #       - ...
+            # TODO: don't distinguish between pk, pk__exaxt, id, id__exact
+            # TOOD: work with .filter(**kwargs).get() ?
             if self._cacheprofile['local_get']        \
                     and not args                      \
                     and not self.query.select_related \
@@ -350,6 +353,16 @@ class QuerySetMixin(object):
 
         return qs._no_monkey.get(qs, *args, **kwargs)
 
+    def first(self):
+        if self._cacheprofile and 'get' in self._cacheprofile['ops']:
+            return self._no_monkey.first(self._clone().cache())
+        return self._no_monkey.first(self)
+
+    def last(self):
+        if self._cacheprofile and 'get' in self._cacheprofile['ops']:
+            return self._no_monkey.last(self._clone().cache())
+        return self._no_monkey.last(self)
+
     def exists(self):
         if self._cacheprofile and 'exists' in self._cacheprofile['ops']:
             if self._result_cache is not None:
@@ -369,10 +382,12 @@ class QuerySetMixin(object):
         clone = self._clone().nocache()
         clone._for_write = True  # affects routing
 
-        objects = list(clone.iterator())  # bypass queryset cache
+        objects = list(clone)
         rows = clone.update(**kwargs)
-        objects.extend(clone.iterator())
-        for obj in objects:
+
+        # TODO: do not refetch objects but update with kwargs in simple cases?
+        pks = {obj.pk for obj in objects}
+        for obj in chain(objects, self.model.objects.filter(pk__in=pks)):
             invalidate_obj(obj)
         return rows
 
@@ -380,7 +395,7 @@ class QuerySetMixin(object):
 def connect_first(signal, receiver, sender):
     old_receivers = signal.receivers
     signal.receivers = []
-    signal.connect(receiver, sender=sender)
+    signal.connect(receiver, sender=sender, weak=False)
     signal.receivers += old_receivers
 
 # We need to stash old object before Model.save() to invalidate on its properties
@@ -389,40 +404,43 @@ _old_objs = threading.local()
 class ManagerMixin(object):
     @once_per('cls')
     def _install_cacheops(self, cls):
-        if family_has_profile(cls):
-            # Set up signals
-            connect_first(pre_save, self._pre_save, sender=cls)
-            connect_first(post_save, self._post_save, sender=cls)
-            connect_first(post_delete, self._post_delete, sender=cls)
+        # Set up signals
+        connect_first(pre_save, self._pre_save, sender=cls)
+        connect_first(post_save, self._post_save, sender=cls)
+        connect_first(post_delete, self._post_delete, sender=cls)
 
-            # Install auto-created models as their module attributes to make them picklable
-            module = sys.modules[cls.__module__]
-            if not hasattr(module, cls.__name__):
-                setattr(module, cls.__name__, cls)
+        # Install auto-created models as their module attributes to make them picklable
+        module = sys.modules[cls.__module__]
+        if not hasattr(module, cls.__name__):
+            setattr(module, cls.__name__, cls)
 
     def contribute_to_class(self, cls, name):
         self._no_monkey.contribute_to_class(self, cls, name)
-        # Django 1.7+ migrations create lots of fake models, just skip them
+        # Django migrations create lots of fake models, just skip them
         # NOTE: we make it here rather then inside _install_cacheops()
         #       because we don't want @once_per() to hold refs to all of them.
-        if not model_is_fake(cls):
+        if cls.__module__ != '__fake__' and family_has_profile(cls):
             self._install_cacheops(cls)
 
     def _pre_save(self, sender, instance, **kwargs):
-        if instance.pk is not None and not no_invalidation.active:
+        if not (instance.pk is None or instance._state.adding or no_invalidation.active):
             try:
                 _old_objs.__dict__[sender, instance.pk] = sender.objects.get(pk=instance.pk)
             except sender.DoesNotExist:
                 pass
 
     def _post_save(self, sender, instance, **kwargs):
+        if not settings.CACHEOPS_ENABLED:
+            return
+
         # Invoke invalidations for both old and new versions of saved object
         old = _old_objs.__dict__.pop((sender, instance.pk), None)
         if old:
             invalidate_obj(old)
         invalidate_obj(instance)
 
-        if transaction_state.is_dirty() or not settings.CACHEOPS_ENABLED:
+        # We run invalidations but skip caching if we are dirty
+        if transaction_states[instance._state.db].is_dirty():
             return
 
         # NOTE: it's possible for this to be a subclass, e.g. proxy, without cacheprofile,
@@ -451,8 +469,6 @@ class ManagerMixin(object):
             key = 'pk' if cache_on_save is True else cache_on_save
             cond = {key: getattr(instance, key)}
             qs = sender.objects.inplace().filter(**cond).order_by()
-            if MAX_GET_RESULTS:
-                qs = qs[:MAX_GET_RESULTS + 1]
             qs._cache_results(qs._cache_key(), [instance])
 
             # Reverting stripped attributes
@@ -491,20 +507,23 @@ def invalidate_m2m(sender=None, instance=None, model=None, action=None, pk_set=N
     if action not in ('pre_clear', 'post_add', 'pre_remove'):
         return
 
+    # NOTE: .rel moved to .remote_field in Django 1.9
+    if django.VERSION >= (1, 9):
+        get_remote = lambda f: f.remote_field
+    else:
+        get_remote = lambda f: f.rel
     m2m = next(m2m for m2m in instance._meta.many_to_many + model._meta.many_to_many
-                   if m2m.rel.through == sender)
+                   if get_remote(m2m).through == sender)
+    instance_column, model_column = m2m.m2m_column_name(), m2m.m2m_reverse_name()
+    if reverse:
+        instance_column, model_column = model_column, instance_column
 
     # TODO: optimize several invalidate_objs/dicts at once
     if action == 'pre_clear':
-        # TODO: always use column names here once Django 1.3 is dropped
-        instance_field = m2m.m2m_reverse_field_name() if reverse else m2m.m2m_field_name()
-        objects = sender.objects.filter(**{instance_field: instance.pk})
+        objects = sender.objects.filter(**{instance_column: instance.pk})
         for obj in objects:
             invalidate_obj(obj)
     elif action in ('post_add', 'pre_remove'):
-        instance_column, model_column = m2m.m2m_column_name(), m2m.m2m_reverse_name()
-        if reverse:
-            instance_column, model_column = model_column, instance_column
         # NOTE: we don't need to query through objects here,
         #       cause we already know all their meaningfull attributes.
         for pk in pk_set:
@@ -521,23 +540,27 @@ def install_cacheops():
     """
     monkey_mix(Manager, ManagerMixin)
     monkey_mix(QuerySet, QuerySetMixin)
-    QuerySet._cacheprofile = QuerySetMixin._cacheprofile
-    QuerySet._cloning = QuerySetMixin._cloning
-
-    # DateQuerySet existed in Django 1.7 and earlier
-    # Values*QuerySet existed in Django 1.8 and earlier
-    from django.db.models import query
-    for cls_name in ('ValuesQuerySet', 'ValuesListQuerySet', 'DateQuerySet'):
-        if hasattr(query, cls_name):
-            cls = getattr(query, cls_name)
-            monkey_mix(cls, QuerySetMixin, ['iterator'])
 
     # Use app registry to introspect used apps
     from django.apps import apps
 
     # Install profile and signal handlers for any earlier created models
     for model in apps.get_models(include_auto_created=True):
-        model._default_manager._install_cacheops(model)
+        if family_has_profile(model):
+            if not isinstance(model._default_manager, Manager):
+                raise ImproperlyConfigured("Can't install cacheops for %s.%s model:"
+                                           " non-django model class or manager is used."
+                                            % (model._meta.app_label, model._meta.model_name))
+            model._default_manager._install_cacheops(model)
+
+            # Bind m2m changed handlers
+            rel_attr = 'remote_field' if django.VERSION >= (1, 9) else 'rel'
+            m2ms = (f for f in model._meta.get_fields(include_hidden=True) if f.many_to_many)
+            for m2m in m2ms:
+                rel = m2m if hasattr(m2m, 'through') else getattr(m2m, rel_attr, m2m)
+                opts = rel.through._meta
+                m2m_changed.connect(invalidate_m2m, sender=rel.through,
+                                    dispatch_uid=(opts.app_label, opts.model_name))
 
     # Turn off caching in admin
     if apps.is_installed('django.contrib.admin'):
@@ -547,9 +570,6 @@ def install_cacheops():
         def get_queryset(self, request):
             return get_queryset.original(self, request).nocache()
 
-    # Bind m2m changed handler
-    m2m_changed.connect(invalidate_m2m)
-
     # Make buffers/memoryviews pickleable to serialize binary field data
     if six.PY2:
         import copy_reg
